import random

Set = frozenset # Data will be frozensets, so they can't be mutated.

def words(text):
    "All space-separated words in text."
    return Set(text.split())

def phrases(text, sep='/'):
    "All sep-separated phrases in text, uppercased and stripped."
    return Set(p.upper().strip() for p in text.split(sep))


winners = words('''washington adams jefferson jefferson madison madison monroe
    monroe adams jackson jackson van-buren harrison polk taylor pierce buchanan
    lincoln lincoln grant grant hayes garfield cleveland harrison cleveland mckinley
    mckinley roosevelt taft wilson wilson harding coolidge hoover roosevelt
    roosevelt roosevelt roosevelt truman eisenhower eisenhower kennedy johnson nixon
    nixon carter reagan reagan bush clinton clinton bush bush obama obama''')
losers = words('''clinton jefferson adams pinckney pinckney clinton king adams
    jackson adams clay van-buren van-buren clay cass scott fremont breckinridge
    mcclellan seymour greeley tilden hancock blaine cleveland harrison bryan bryan
    parker bryan roosevelt hughes cox davis smith hoover landon wilkie dewey dewey
    stevenson stevenson nixon goldwater humphrey mcgovern ford carter mondale
    dukakis bush dole gore kerry mccain romney''') - winners

boys = words('jacob mason ethan noah william liam jayden michael alexander aiden')
girls = words('sophia emma isabella olivia ava emily abigail mia madison elizabeth')

pharma = words('lipitor nexium plavix advair ablify seroquel singulair crestor actos epogen')
cities = words('paris trinidad capetown riga zurich shanghai vancouver chicago adelaide auckland')

foo = words('''afoot catfoot dogfoot fanfoot foody foolery foolish fooster footage foothot footle footpad footway
    hotfoot jawfoot mafoo nonfood padfoot prefool sfoot unfool''')
bar = words('''Atlas Aymoro Iberic Mahran Ormazd Silipan altared chandoo crenel crooked fardo folksy forest
    hebamic idgah manlike marly palazzi sixfold tarrock unfold''')

nouns = words('''time year people way day man thing woman life child world school
    state family student group country problem hand part place case week company
    system program question work government number night point home water room
    mother area money story fact month lot right study book eye job word business
    issue side kind head house service friend father power hour game line end member
    law car city community name president team minute idea kid body information
    back parent face others level office door health person art war history party result
    change morning reason research girl guy moment air teacher force education''')
adverbs = words('''all particularly just less indeed over soon course still yet before
    certainly how actually better to finally pretty then around very early nearly now
    always either where right often hard back home best out even away enough probably
    ever recently never however here quite alone both about ok ahead of usually already
    suddenly down simply long directly little fast there only least quickly much forward
    today more on exactly else up sometimes eventually almost thus tonight as in close
    clearly again no perhaps that when also instead really most why ago off
    especially maybe later well together rather so far once''') - nouns
verbs = words('''ask believe borrow break bring buy can be able cancel change clean
    comb complain cough count cut dance draw drink drive eat explain fall
    fill find finish fit fix fly forget give go have hear hurt know learn
    leave listen live look lose make do need open close shut organise pay
    play put rain read reply run say see sell send sign sing sit sleep
    smoke speak spell spend stand start begin study succeed swim take talk
    teach tell think translate travel try turn off turn on type understand
    use wait wake up want watch work worry write''') - nouns

# randoms = Set(vars(random))
# print(randoms)
# builtins = Set(vars(__builtin__)) - randoms

starwars = phrases('''The Phantom Menace / Attack of the Clones / Revenge of the Sith /
    A New Hope / The Empire Strikes Back / Return of the Jedi''')
startrek = phrases('''The Wrath of Khan / The Search for Spock / The Voyage Home /
    The Final Frontier / The Undiscovered Country / Generations / First Contact /
    Insurrection / Nemesis''')

dogs = phrases(''''Labrador Retrievers / German Shepherd Dogs / Golden Retrievers / Beagles / Bulldogs /
    Yorkshire Terriers / Boxers / Poodles / Rottweilers / Dachshunds / Shih Tzu / Doberman Pinschers /
    Miniature Schnauzers / French Bulldogs / German Shorthaired Pointers / Siberian Huskies / Great Danes /
    Chihuahuas / Pomeranians / Cavalier King Charles Spaniels / Shetland Sheepdogs / Australian Shepherds /
    Boston Terriers / Pembroke Welsh Corgis / Maltese / Mastiffs / Cocker Spaniels / Havanese /
    English Springer Spaniels / Pugs / Brittanys / Weimaraners / Bernese Mountain Dogs / Vizslas / Collies /
    West Highland White Terriers / Papillons / Bichons Frises / Bullmastiffs / Basset Hounds /
    Rhodesian Ridgebacks / Newfoundlands / Russell Terriers / Border Collies / Akitas /
    Chesapeake Bay Retrievers / Miniature Pinschers / Bloodhounds / St. Bernards / Shiba Inu / Bull Terriers /
    Chinese Shar-Pei / Soft Coated Wheaten Terriers / Airedale Terriers / Portuguese Water Dogs / Whippets /
    Alaskan Malamutes / Scottish Terriers / Australian Cattle Dogs / Cane Corso / Lhasa Apsos /
    Chinese Crested / Cairn Terriers / English Cocker Spaniels / Dalmatians / Italian Greyhounds /
    Dogues de Bordeaux / Samoyeds / Chow Chows / German Wirehaired Pointers / Belgian Malinois /
    Great Pyrenees / Pekingese / Irish Setters / Cardigan Welsh Corgis / Staffordshire Bull Terriers /
    Irish Wolfhounds / Old English Sheepdogs / American Staffordshire Terriers / Bouviers des Flandres /
    Greater Swiss Mountain Dogs / Japanese Chin / Tibetan Terriers / Brussels Griffons /
    Wirehaired Pointing Griffons / Border Terriers / English Setters / Basenjis / Standard Schnauzers /
    Silky Terriers / Flat-Coated Retrievers / Norwich Terriers / Afghan Hounds / Giant Schnauzers / Borzois /
    Wire Fox Terriers / Parson Russell Terriers / Schipperkes / Gordon Setters / Treeing Walker Coonhounds''')
cats = phrases('''Abyssinian / Aegean cat / Australian Mist / American Curl / American Bobtail /
    American Polydactyl / American Shorthair / American Wirehair / Arabian Mau / Asian / Asian Semi-longhair /
    Balinese / Bambino / Bengal / Birman / Bombay / Brazilian Shorthair / British Shorthair / British Longhair /
    Burmese / Burmilla / California Spangled Cat / Chantilly/Tiffany / Chartreux / Chausie / Cheetoh /
    Colorpoint Shorthair / Cornish Rex / Cymric / Cyprus cat / Devon Rex / Donskoy or Don Sphynx / Dragon Li /
    Dwelf / Egyptian Mau / European Shorthair / Exotic Shorthair / German Rex / Havana Brown / Highlander /
    Himalayan-Colorpoint Persian / Japanese Bobtail / Javanese / Khao Manee / Korat / Korn Ja /
    Kurilian Bobtail / LaPerm / Maine Coon / Manx / Mekong bobtail / Minskin / Munchkin / Nebelung / Napoleon /
    Norwegian Forest Cat / Ocicat / Ojos Azules / Oregon Rex / Oriental Bicolor / Oriental Shorthair /
    Oriental Longhair / Persian / Peterbald / Pixie-bob / Ragamuffin / Ragdoll / Russian Blue / Russian Black /
    Sam Sawet / Savannah / Scottish Fold / Selkirk Rex / Serengeti cat / Serrade petit / Siamese / Siberian /
    Singapura / Snowshoe / Sokoke / Somali / Sphynx / Swedish forest cat / Thai / Tonkinese / Toyger /
    Turkish Angora / Turkish Van / Ukrainian Levkoy / York Chocolate Cat''')

movies = phrases('''Citizen Kane / The Godfather / Vertigo / 2001: A Space Odyssey / The Searchers / Sunrise /
    Singin’ in the Rain / Psycho / Casablanca / The Godfather Part II / The Magnificent Ambersons / Chinatown /
    North by Northwest / Nashville / The Best Years of Our Lives / McCabe & Mrs Miller / The Gold Rush /
    City Lights / Taxi Driver / Goodfellas / Mulholland Drive / Greed / Annie Hall / The Apartment /
    Do the Right Thing / Killer of Sheep / Barry Lyndon / Pulp Fiction / Raging Bull / Some Like It Hot /
    A Woman Under the Influence / The Lady Eve / The Conversation / The Wizard of Oz / Double Indemnity /
    Star Wars / Imitation of Life / Jaws / The Birth of a Nation / Meshes of the Afternoon / Rio Bravo /
    Dr Strangelove / Letter from an Unknown Woman / Sherlock Jr / The Man Who Shot Liberty Valance /
    It’s a Wonderful Life / Marnie / A Place in the Sun / Days of Heaven / His Girl Friday / Touch of Evil /
    The Wild Bunch / Grey Gardens / Sunset Boulevard / The Graduate / Back to the Future / Crimes and Misdemeanors /
    The Shop Around the Corner / One Flew Over the Cuckoo’s Nest / Blue Velvet / Eyes Wide Shut / The Shining /
    Love Streams / Johnny Guitar / The Right Stuff / Red River / Modern Times / Notorious / Koyaanisqatsi /
    The Band Wagon / Groundhog Day / The Shanghai Gesture / Network / Forrest Gump /
    Close Encounters of the Third Kind / The Empire Strikes Back / Stagecoach / Schindler’s List /
    The Tree of Life / Meet Me in St Louis / Thelma & Louise / Raiders of the Lost Ark / Bringing Up Baby /
    Deliverance / Night of the Living Dead / The Lion King / Eternal Sunshine of the Spotless Mind /
    West Side Story / In a Lonely Place / Apocalypse Now / ET: The Extra-Terrestrial / The Night of the Hunter /
    Mean Streets / 25th Hour / Duck Soup / The Dark Knight / Gone With the Wind / Heaven’s Gate / 12 Years a Slave /
    Ace in the Hole''')
tv = phrases('''The Abbott and Costello Show / ABC’s Wide World of Sports / Alfred Hitchcock Presents /
    All in the Family / An American Family / American Idol / Arrested Development / Battlestar Galactica /
    The Beavis and Butt-Head Show / The Bob Newhart Show / Brideshead Revisited / Buffalo Bill /
    Buffy the Vampire Slayer / The Carol Burnett Show / The CBS Evening News with Walter Cronkite /
    A Charlie Brown Christmas / Cheers / The Cosby Show / The Daily Show / Dallas / The Day After /
    Deadwood / The Dick Van Dyke Show / Dragnet / The Ed Sullivan Show / The Ernie Kovacs Show /
    Felicity / Freaks and Geeks / The French Chef / Friends / General Hospital /
    The George Burns and Gracie Allen Show / Gilmore Girls / Gunsmoke / Hill Street Blues /
    Homicide: Life on the Street / The Honeymooners / I, Claudius / I Love Lucy / King of the Hill /
    The Larry Sanders Show / Late Night with David Letterman / Leave It to Beaver / Lost /
    Married with Children / Mary Hartman, Mary Hartman / The Mary Tyler Moore Show / MASH / The Monkees /
    Monty Python’s Flying Circus / Moonlighting / My So-Called Life / Mystery Science Theater 3000 /
    The Odd Couple / The Office / The Oprah Winfrey Show / Pee Wee’s Playhouse / Playhouse 90 /
    The Price Is Right / Prime Suspect / The Prisoner / The Real World / Rocky and His Friends / Roots /
    Roseanne / Sanford and Son / Saturday Night Live / Second City Television / See It Now / Seinfeld /
    Sesame Street / Sex and the City / The Shield / The Simpsons / The Singing Detective / Six Feet Under /
    60 Minutes / Soap / The Sopranos / South Park / SpongeBob SquarePants / SportsCenter / Star Trek /
    St Elsewhere / The Super Bowl / Survivor / Taxi /The Tonight Show Starring Johnny Carson /
    24 / The Twilight Zone / Twin Peaks / The West Wing / What’s My Line / WKRP in Cincinnati /
    The Wire / Wiseguy / The X-Files''')

stars = phrases('''Humphrey Bogart / Cary Grant / James Stewart / Marlon Brando / Fred Astaire / Henry Fonda /
    Clark Gable / James Cagney / Spencer Tracy / Charlie Chaplin / Gary Cooper / Gregory Peck / John Wayne /
    Laurence Olivier / Gene Kelly / Orson Welles / Kirk Douglas / James Dean / Burt Lancaster / Marx Brothers /
    Buster Keaton / Sidney Poitier / Robert Mitchum / Edward G. Robinson / William Holden / Katharine Hepburn /
    Bette Davis / Audrey Hepburn / Ingrid Bergman / Greta Garbo / Marilyn Monroe / Elizabeth Taylor / Judy Garland /
    Marlene Dietrich / Joan Crawford / Barbara Stanwyck / Claudette Colbert / Grace Kelly / Ginger Rogers /
    Mae West / Vivien Leigh / Lillian Gish / Shirley Temple / Rita Hayworth / Lauren Bacall / Sophia Loren /
    Jean Harlow / Carole Lombard / Mary Pickford / Ava Gardner''')
scientists = phrases('''Alain Aspect / Martin Karplus / David Baltimore / Donald Knuth / Allen Bard /
    Robert Marks II / Timothy Berners-Lee / Craig Mello / John Tyler Bonner / Luc Montagnier / Dennis Bray /
    Gordon Moore / Sydney Brenner / Kary Mullis / Pierre Chambon / C Nüsslein-Volhard / Simon Conway Morris /
    Seiji Ogawa / Mildred Dresselhaus / Jeremiah Ostriker / Gerald M Edelman / Roger Penrose / Ronald Evans /
    Stanley Prusiner / Anthony Fauci / Henry F Schaefer III / Anthony Fire / Thomas Südhof / Jean Frechet /
    Jack Szostak / Margaret Geller / James Tour / Jane Goodall / Charles Townes / Alan Guth / Harold Varmus /
    Lene Vestergaard Hau / Craig Venter / Stephen Hawking / James Watson / Peter Higgs / Steven Weinberg /
    Leroy Hood / George Whitesides / Eric Kandel / Edward Wilson / Andrew Knoll / Edward Witten / Charles Kao /
    Shinya Yamanaka''')


_winners = (winners, boys, pharma, foo, nouns, starwars, dogs, movies, stars)
_losers = (losers, girls, cities, bar, adverbs, startrek, cats, tv, scientists)



from .REZRegex import REZRegex
from ..base.interface import *

# I can't figure out how flags work in R, so I'm just ignoring them
del line_starts_with
del lineStartsWith
del line_start
del lineStart
del line_ends_with
del lineEndsWith
del line_end
del lineEnd

del ASCII
del DOTALL
del IGNORECASE
del LOCALE
del MULTILINE
del UNICODE

"Group: Not Literals"
not_newline: REZRegex = REZRegex(r'\N')
'Matches any single character except line break characters, like the dot, but is not affected by any options that make the dot match all characters including line breaks.'



from ..EZRegex import EZRegex
import re

class REZRegex(EZRegex):
    # Source: https://www.rdocumentation.org/packages/base/versions/3.6.2/topics/regex
    _escape_chars=b'|()[{^$*+?.-'
    _repl_escape_chars=b''

    def _final_func(self, s:str) -> str:
        # Double escape all backslashes
        s = s.replace('\\', '\\\\')
        # Things which are control characters should *not* be double escaped
        s = re.sub(r'\\\\([abtnvfrxuU])', r'\\\g<1>', s)
        return s



import json
import string
import traceback
from pathlib import Path
from random import choice, choices, randint
from re import search
from sys import version_info
from typing import Literal, Union
import logging

from ezregex import *

if version_info.minor <= 10:
    from re import sre_parse as sre  # type: ignore
else:
    from re import _parser as sre  # type: ignore

# So I can debug this function directly
if __name__ != '__main__':
    from .invert_old import invertRegex
else:
    from ezregex.invert_old import invertRegex

# TODO: add an option to put the matching string inside stuff that doesn't match
# TODO: add an option to include multiple matching strings inside stuff that doesn't match

with open(Path(__file__).parent / 'assets' / 'common_sorted_words.json') as f:
    words = json.load(f)


def invert(
    expr:Union[str, 'EZRegex'],
    tries:int=10,
    backend:Literal['re_parser', 'regex', 'xeger', 'sre_yield']=...,
    words:Literal['lookup', 'random']|None='lookup',
    randomNumbers=False,
    alot=8,
) -> str:
    return Inverter(expr,tries, backend, words, randomNumbers, alot).invert()


class Inverter:
    """ "Inverts" a regular expression by returning an example of something which is guaruanteed to
        match the passed expression.
        NOTE: This only works on valid Python regular expressions.

        Args:
            expr: The regular expression to invert. Can be a string, or a EZRegex expression
            words: Controls how works are handled. If `random`, words are made of random letters. If `lookup`,
                it looks up valid english words and inserts them to make it more readable.
            randomNumbers: controls whether all numbers are 12345... to a desired length, or if they're
                just random numbers (again, for readability)
            self.alot: When given a choice of how many characters to put someone, it inserts a random integer
                between 1 and `alot`.
            tries: Controls how many times to try to invert the expression before giving up. This is effective,
                because there is an element of randomness involved in inverting the regex given.
            backend: One of ('re_parser', 'regex', 'xeger', 'sre_yield').
                `re_parser` is the default, it uses the build-in parser in the re package to create an AST
                    of the regex
                `regex` uses regular expressions to parse regular expressions, which is as gross as it sounds.
                    It's slightly buggy, but mostly works.
                `xeger` imports the `xeger` pacakge and uses it instead. Xeger inverts work, but are less
                    readable. Must have the `xeger` package installed.
                `sre_yield` imports the `sre_yield` package and uses it instead. I don't think this works
                    right now, and may be significantly slower.
    """
    def __init__(self,
        expr:Union[str, 'EZRegex'],
        tries:int=10,
        backend:Literal['re_parser', 'regex', 'xeger', 'sre_yield']=...,
        words:Literal['lookup', 'random']|None='lookup',
        randomNumbers=False,
        alot=8,
    ):
        self.expr = str(expr)
        self.words = words
        self.randomNumbers = randomNumbers
        self.alot = alot
        self.tries = tries
        self.backend = backend
        self._attempts = {
            're_parser': 0,
            'regex': 0,
            'xeger': 0,
            'sre_yield': 0,
        }
        # Don't use string.whitespace, because we don't want to use weird difficult to print characters.
        # We want the replacement to be readable.
        self._whitespace = ' '
        self._everything = string.digits + string.ascii_letters + string.punctuation + self._whitespace + '_'

        try: from xeger import Xeger  # type: ignore
        except ImportError: self._xeger = False
        else: self._xeger = True

        try: import sre_yield  # type: ignore
        except ImportError: self._sre_yield = False
        else: self._sre_yield = True

        if backend == 'xeger' and not self._xeger:
            raise ImportError(f'Requested backend `xeger` not available. Try installing it by running `pip install xeger`')

        if backend == 'sre_yield' and not self._sre_yield:
            raise ImportError(f'Requested backend `sre_yield` not available. Try installing it by running `pip install sre_yield`')

    def _randWord(self, length=..., word=...) -> str:
        if word is Ellipsis:
            word = self.words
        if length is Ellipsis:
            length = randint(3, self.alot)

        if word == 'random':
            return ''.join(choices(string.ascii_letters, k=length))

        elif word == 'lookup':
            try:
                # The keys are strings, not ints, and I don't feel like fixing it
                return choice(words[str(length)])
            except KeyError:
                return self._randWord(length, word='random')

        elif word is None:
            return 'word'

        else:
            raise ValueError(f"invalid parameter given for word {word}. Accepted values are either random, lookup, or None")

    def _randNumber(self, length=...) -> str:
        if length is Ellipsis:
            length = randint(2, self.alot)

        if self.randomNumbers:
            return str(randint(0, 10**length))
        else:
            return ''.join(map(lambda i: str(i)[-1], range(1, length+1)))

    def invert_re_parser(self) -> str | None:
        self._attempts['re_parser'] += 1
        logging.debug(f're_parser attempt #{self._attempts["re_parser"]}...')
        groups = {}
        def handle(pattern, amt=1, opposite=False):
            logging.debug(f'Handling {pattern} * {amt}')
            s = ''
            for op, args in pattern:
                if not opposite:
                    match op:
                        case sre.LITERAL:
                            s += chr(args) * amt
                        case sre.NOT_LITERAL:
                            almost_everything = list(self._everything)
                            if chr(args) in almost_everything:
                                almost_everything.remove(chr(args))
                            # If it's not in there, great!
                            s += choice(almost_everything) * amt
                        case sre.RANGE:
                            start, end = args
                            s += chr(randint(start, end))
                        case sre.MIN_REPEAT: # optional, I believe
                            min, max, sub = args
                            if randint(0, 1) == 1:
                                s += handle(sub, amt)
                        case sre.MAX_REPEAT:
                            min, max, sub = args
                            if max is None or max is sre.MAXREPEAT:
                                max = randint(min, self.alot)
                            try:
                                if type(sub[0][1]) is int:
                                    s += handle(sub, amt)
                                # If we know we're getting word chars, make a word of it
                                elif type(sub[0][1][0]) is not int and sub[0][1][0][1] in (sre.CATEGORY_WORD, sre.CATEGORY_UNI_WORD, sre.CATEGORY_LOC_WORD):
                                    logging.debug('Getting a random word')
                                    s += self._randWord()
                                # If we know we're getting digit chars, make a number of it
                                elif type(sub[0][1][0]) is not int and sub[0][1][0][1] in (sre.CATEGORY_DIGIT, sre.CATEGORY_UNI_DIGIT):
                                    logging.debug('Getting whole number')
                                    s += self._randNumber()
                                else:
                                    s += handle(sub, randint(min, max) * amt)
                            except Exception as err:
                                logging.debug(err)
                                logging.debug(traceback.format_exc())

                                s += handle(sub, randint(min, max) * amt)
                        case sre.ANY:
                            for _ in range(amt):
                                s += choice(self._everything)
                        case sre.ASSERT:
                            type_, sub = args
                            if type_ == 1: # ifProcededBy
                                s = s + handle(sub, amt)
                            elif type_ == -1: # ifPrecededBy
                                s += handle(sub, amt)
                            else:
                                s += handle(sub, amt)
                        case sre.ASSERT_NOT:
                            type_, sub = args
                            if type_ == 1: # ifNotProcededBy
                                # almost_everything = list(self._everything)
                                # if chr(args) in almost_everything:
                                #     almost_everything.remove(chr(args))
                                # # If it's not in there, great!
                                # s += choice(almost_everything) * amt
                                s += handle(sub, amt, opposite=True)
                            if type_ == -1: # ifNotPrecededBy
                                s = s + handle(sub, amt, opposite=True)
                            # If it needs to be followed by that sequence, simply add that sequence
                            else:
                                s += handle(sub, amt, opposite=True)
                                # s += handle(sub, amt)
                        case sre.CATEGORY:
                            for _ in range(amt):
                                match args:
                                    case sre.CATEGORY_DIGIT | sre.CATEGORY_UNI_DIGIT:
                                        s += str(randint(0, 10))
                                    case sre.CATEGORY_LINEBREAK | sre.CATEGORY_UNI_LINEBREAK:
                                        s += '\n'
                                    case sre.CATEGORY_NOT_DIGIT | sre.CATEGORY_UNI_NOT_DIGIT:
                                        s += choice(string.ascii_letters + string.punctuation + self._whitespace)
                                    case sre.CATEGORY_NOT_LINEBREAK | sre.CATEGORY_UNI_NOT_LINEBREAK:
                                        s += choice(self._everything)
                                    case sre.CATEGORY_NOT_SPACE | sre.CATEGORY_UNI_NOT_SPACE:
                                        s += choice(string.punctuation + string.ascii_letters + string.digits)
                                    case sre.CATEGORY_NOT_WORD | sre.CATEGORY_UNI_NOT_WORD | sre.CATEGORY_LOC_NOT_WORD:
                                        s += choice(string.punctuation + self._whitespace)
                                    case sre.CATEGORY_SPACE | sre.CATEGORY_UNI_SPACE:
                                        s += ' '
                                    case sre.CATEGORY_WORD | sre.CATEGORY_UNI_WORD | sre.CATEGORY_LOC_WORD:
                                        s += choice(string.ascii_letters + '_')
                                    case _:
                                        raise NotImplementedError(f'Unknown category given: {args}')
                        case sre.IN:
                            # If this is a pattern like [^abc]
                            if args[0][0] is sre.NEGATE:
                                # Handle all the args except the negate flag and remove it from the options
                                otherthan = handle(args[1:])
                                almost_everything = list(self._everything)
                                for i in otherthan:
                                    if i in almost_everything:
                                        almost_everything.remove(i)
                                s += choice(almost_everything)
                            else:
                                s += handle([choice(args)], amt)
                        case sre.SUBPATTERN: # This handles groups
                            group, num, num2, sub = args
                            sub_pattern = handle(sub) * amt
                            groups[group] = sub_pattern
                            s += sub_pattern
                        case sre.AT:
                            if args is sre.AT_BEGINNING_STRING:
                                # Whatever comes next better be first
                                s = ''
                            elif args is sre.AT_END_STRING:
                                # Whatever comes next better be last
                                return s
                            elif args is sre.AT_BEGINNING:
                                s += '\n'
                            elif args is sre.AT_END:
                                s += '\n'
                            elif args is sre.AT_BOUNDARY:
                                if len(s) and s[-1] in string.digits + string.ascii_letters + '_':
                                    s += choice(string.punctuation + self._whitespace)
                                else:
                                    s += choice(string.digits + string.ascii_letters + '_')
                            # \B, I believe
                            elif args is sre.AT_NON_BOUNDARY:
                                # This works... I'm not entirely sure why... I'm not gonna touch it.
                                if len(s) and s[-1] not in string.digits + string.ascii_letters + '_':
                                    s += choice(string.digits + string.ascii_letters + '_')
                                else:
                                    s += choice(string.punctuation + self._whitespace)
                            else:
                                raise NotImplementedError(f'Unknown parameter[s] given for AT op: {args}')
                        case sre.BRANCH:
                            something, branches = args
                            s += handle(choice(branches), amt)
                        case sre.NEGATE:
                            s += handle(args, amt, opposite=True)
                        case sre.GROUPREF:
                            s += groups[args]
                        case sre.GROUPREF_EXISTS:
                            # TODO: This often requires multiple tries to get right
                            group, trueSub, falseSub = args
                            s += handle(trueSub if group in groups else falseSub, amt)
                        case _:
                            raise NotImplementedError(f'Unknown op {op} given with args {args}')
                else:
                    match op:
                        case sre.LITERAL:
                            almost_everything = list(self._everything)
                            if chr(args) in almost_everything:
                                almost_everything.remove(chr(args))
                            # If it's not in there, great!
                            s += choice(almost_everything) * amt
                        case sre.NOT_LITERAL:
                            s += chr(args) * amt
                        case sre.RANGE:
                            start, end = args
                            almost_everything = set(self._everything)
                            almost_everything.difference(map(chr, range(start, end)))
                            s += choice(list(almost_everything))
                        case sre.MAX_REPEAT | sre.MIN_REPEAT:
                            min, max, sub = args
                            # Adding something that wouldn't match at all is always an option
                            options = [handle(sub, amt, opposite=True)]
                            # If a max is specified, then it can't match more than that many
                            if max is not None and max is not sre.MAXREPEAT:
                                options.append(handle(sub, amt * (max + randint(1, self.alot))))
                            # If there's a min (meaning its more than 0), then it can't match
                            # less than that many
                            if min >= 1:
                                # TODO: Potential Error: This doesn't take into account the current amt.
                                options.append(handle(sub, randint(0, min-1)))
                            s += choice(options)
                        case sre.ANY:
                            # TODO: I think this will fail if given [^.\\n] (or anyExcept(literallyAnything)),
                            # but also, how DO you handle that?
                            for _ in range(amt):
                                # I guess?
                                s += '\n'
                        case sre.ASSERT:
                            # I don't think conditionals are allowed inside a not assert (I think)
                            type_, sub = args
                            s += handle(sub, amt, opposite=True)
                        case sre.ASSERT_NOT:
                            type_, sub = args
                            # I *think* this will work?...
                            s += handle(sub)
                        case sre.CATEGORY:
                            for _ in range(amt):
                                match args:
                                    case sre.CATEGORY_DIGIT | sre.CATEGORY_UNI_DIGIT:
                                        s += choice(string.ascii_letters + string.punctuation + self._whitespace)
                                    case sre.CATEGORY_LINEBREAK | sre.CATEGORY_UNI_LINEBREAK:
                                        s += choice(self._everything)
                                    case sre.CATEGORY_NOT_DIGIT | sre.CATEGORY_UNI_NOT_DIGIT:
                                        s += choice(string.digits)
                                    case sre.CATEGORY_NOT_LINEBREAK | sre.CATEGORY_UNI_NOT_LINEBREAK:
                                        s += '\n'
                                    case sre.CATEGORY_NOT_SPACE | sre.CATEGORY_UNI_NOT_SPACE:
                                        s += ' '
                                    case sre.CATEGORY_NOT_WORD | sre.CATEGORY_UNI_NOT_WORD | sre.CATEGORY_LOC_NOT_WORD:
                                        s += choice(string.ascii_letters + '_')
                                    case sre.CATEGORY_SPACE | sre.CATEGORY_UNI_SPACE:
                                        s += choice(string.punctuation + string.ascii_letters + string.digits)
                                    case sre.CATEGORY_WORD | sre.CATEGORY_UNI_WORD | sre.CATEGORY_LOC_WORD:
                                        s += choice(string.punctuation + self._whitespace)
                                    case _:
                                        raise NotImplementedError(f'Unknown category given: {args}')
                        case sre.IN:
                            # If this is a pattern like [^abc]
                            if args[0][0] is not sre.NEGATE:
                                # Handle all the args except the negate flag and remove it from the options
                                otherthan = handle(args[1:])
                                almost_everything = list(self._everything)
                                for i in otherthan:
                                    if i in almost_everything:
                                        almost_everything.remove(i)
                                s += choice(almost_everything)
                            else:
                                s += handle([choice(args)], amt)
                        case sre.SUBPATTERN:
                            idk, num, num2, sub = args
                            s += handle(sub, opposite=True) * randint(0, amt)
                        case None: #sre.AT: # I don't know how to handle this
                            if args is sre.AT_BEGINNING_STRING:
                                # Whatever comes next better be first
                                s = ''
                            elif args is sre.AT_END_STRING:
                                # Whatever comes next better be last
                                return s
                            elif args is sre.AT_BEGINNING:
                                s += '\n'
                            elif args is sre.AT_END:
                                s += '\n'
                            elif args is sre.AT_BOUNDARY:
                                if len(s) and s[-1] in string.digits + string.ascii_letters + '_':
                                    s += choice(string.punctuation + self._whitespace)
                                else:
                                    s += choice(string.digits + string.ascii_letters + '_')
                            else:
                                raise NotImplementedError(f'Unknown parameter[s] given for AT op: {args}')
                        case sre.BRANCH:
                            something, branches = args
                            s += handle(choice(branches), amt, opposite=True)
                        case sre.NEGATE:
                            s += handle(args, amt)
                        case sre.GROUPREF:
                            add = groups[args]
                            while add == groups[args]:
                                add = self._randWord()
                            s += add
                        case sre.GROUPREF_EXISTS:
                            group, trueSub, falseSub = args
                            s += handle(trueSub if group in groups else falseSub, amt, opposite=True)
                        case _:
                            raise NotImplementedError(f'Unknown opposite op {op} given with args {args}')
            return s

        return handle(sre.parse(self.expr))

    def invert_regex(self) -> str | None:
        self._attempts['regex'] += 1
        logging.debug(f'regex attempt #{self._attempts["regex"]}...')
        try:
            rtn = invertRegex(self.expr, 1)
        except NotImplementedError:
            pass

    def invert_xeger(self) -> str | None:
        self._attempts['xeger'] += 1
        logging.debug(f'xeger attempt #{self._attempts["xeger"]}...')
        if self._xeger:
            from xeger import Xeger  # type: ignore
            return Xeger().xeger(self.expr)

    def invert_sre_yield(self) -> str | None:
        self._attempts['sre_yield'] += 1
        logging.debug(f'sre_yield attempt #{self._attempts["sre_yield"]}...')
        if self._sre_yield:
            import sre_yield  # type: ignore
            for i in sre_yield.AllStrings(self.expr):
                return i

    def invert(self) -> str:
        if self.backend is Ellipsis:
            order = ('re_parser', 'regex', 'sre_yield', 'xeger')
            for backend in order:
                while self._attempts[backend] <= self.tries:
                    rtn = getattr(self, 'invert_' + backend)()
                    if rtn is not None and search(self.expr, rtn):
                        logging.info(f'Found using {backend}')
                        return rtn
                    else:
                        if rtn is not None:
                            logging.info(f'Successfully inverted pattern, but it was invalid: `{rtn}`')
                        else:
                            logging.info('Failed to invert pattern')

            logging.info(f'Not found using {backend}')
            msg = (
                f"Failed to invert pattern `{self.expr}`. Likely, bad regex was given. "
                "If you think that's not the case, please submit a bug report to "
                "https://github.com/smartycope/ezregex/issues, and include the regex you used to get this error. "
            )
            if not self._xeger or not self._sre_yield:
                msg += "You can also try installing the extra backends to see if those work:\n"
            if not self._xeger:
                msg += "xeger (`pip install xeger`)\n"
            if not self._sre_yield:
                msg += "sre_yield (`pip install sre-yield`)"

            raise NotImplementedError(msg)
        else:
            while self._attempts[self.backend] <= self.tries:
                rtn = getattr(self, 'invert_' + self.backend)()
                if rtn is not None and search(self.expr, rtn):
                    return rtn
            raise NotImplementedError(f"Failed to invert pattern `{self.expr}` using the `{self.backend}` backend")


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    # inv = Inverter(r'\w+\d+', verbose=False)
    # print(inv.invert())
    # print(invert(r'D(?=AB)C', verbose=True))  # ASSERT,      1, ifProcededBy
    # print(invert(r'D(?!AB)C'))  # ASSERT_NOT,  1, ifNotProcededBy
    # print(invert(r'C(?<=AB)D', _verbose=True)) # ASSERT,     -1, ifPrecededBy
    # print(invert(r'C(?<!AB)D'))   # ASSERT_NOT, -1, ifNotPrecededBy

    # print(invert(r'C(?=AB)'))  # ASSERT,      1, ifProcededBy
    # print(invert(r'C(?!AB)'))  # ASSERT_NOT,  1, ifNotProcededBy
    # print(invert(r'(?<=AB)C')) # ASSERT,     -1, ifPrecededBye
    # print(invert(r'(?<!AB)C'))   # ASSERT_NOT, -1, ifNotPrecededBy

    # print(invert(r'(?=AB)(?!CD)DC AB(?<=CD) AB(?<!CD)'))

    # print(invert(r'[ABC]+(?=D).*$ <.*?>'))
    print(invert(r'(<)?(\w+@\w+(?:\.\w+)+)(?(1)>|$)'))
    print(invert(r'(?:(<))?(\w+@\w+(?:\.\w+)+)(?(1)>|\Z)'))

    # print(invert(r'\w+test\d+', _verbose=True))
    # TODO: I think this will fail if given [^.\\n] (or anyExcept(literallyAnything)),
